@@ -6,44 +6,29 @@
 
 #include "FetchStreamReader.h"
 #include "InternalResponse.h"
+#include "js/Stream.h"
+#include "mozilla/ConsoleReportCollector.h"
+#include "mozilla/ErrorResult.h"
+#include "mozilla/dom/AutoEntryScript.h"
+#include "mozilla/dom/DOMException.h"
+#include "mozilla/dom/Promise.h"
 #include "mozilla/dom/PromiseBinding.h"
-#include "mozilla/SystemGroup.h"
+#include "mozilla/dom/ReadableStream.h"
+#include "mozilla/dom/ReadableStreamDefaultController.h"
+#include "mozilla/dom/ReadableStreamDefaultReader.h"
+#include "mozilla/dom/WorkerPrivate.h"
+#include "mozilla/dom/WorkerRef.h"
+#include "mozilla/HoldDropJSObjects.h"
 #include "mozilla/TaskCategory.h"
 #include "nsContentUtils.h"
+#include "nsDebug.h"
+#include "nsIAsyncInputStream.h"
+#include "nsIPipe.h"
 #include "nsIScriptError.h"
 #include "nsPIDOMWindow.h"
+#include "jsapi.h"
 
-namespace mozilla {
-namespace dom {
-
-namespace {
-
-class FetchStreamReaderWorkerHolder final : public WorkerHolder {
- public:
-  explicit FetchStreamReaderWorkerHolder(FetchStreamReader* aReader)
-      : WorkerHolder("FetchStreamReaderWorkerHolder",
-                     WorkerHolder::Behavior::AllowIdleShutdownStart),
-        mReader(aReader),
-        mWasNotified(false) {}
-
-  bool Notify(WorkerStatus aStatus) override {
-    if (!mWasNotified) {
-      mWasNotified = true;
-      // The WorkerPrivate does have a context available, and we could pass it
-      // here to trigger cancellation of the reader, but the author of this
-      // comment chickened out.
-      mReader->CloseAndRelease(nullptr, NS_ERROR_DOM_INVALID_STATE_ERR);
-    }
-
-    return true;
-  }
-
- private:
-  RefPtr<FetchStreamReader> mReader;
-  bool mWasNotified;
-};
-
-}  // namespace
+namespace mozilla::dom {
 
 NS_IMPL_CYCLE_COLLECTING_ADDREF(FetchStreamReader)
 NS_IMPL_CYCLE_COLLECTING_RELEASE(FetchStreamReader)
@@ -52,14 +37,15 @@ NS_IMPL_CYCLE_COLLECTION_CLASS(FetchStreamReader)
 
 NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN(FetchStreamReader)
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mGlobal)
+  NS_IMPL_CYCLE_COLLECTION_UNLINK(mReader)
 NS_IMPL_CYCLE_COLLECTION_UNLINK_END
 
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_BEGIN(FetchStreamReader)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mGlobal)
+  NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mReader)
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
 
 NS_IMPL_CYCLE_COLLECTION_TRACE_BEGIN(FetchStreamReader)
-  NS_IMPL_CYCLE_COLLECTION_TRACE_JS_MEMBER_CALLBACK(mReader)
 NS_IMPL_CYCLE_COLLECTION_TRACE_END
 
 NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(FetchStreamReader)
@@ -67,9 +53,10 @@ NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(FetchStreamReader)
   NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIOutputStreamCallback)
 NS_INTERFACE_MAP_END
 
-/* static */ nsresult FetchStreamReader::Create(
-    JSContext* aCx, nsIGlobalObject* aGlobal, FetchStreamReader** aStreamReader,
-    nsIInputStream** aInputStream) {
+/* static */
+nsresult FetchStreamReader::Create(JSContext* aCx, nsIGlobalObject* aGlobal,
+                                   FetchStreamReader** aStreamReader,
+                                   nsIInputStream** aInputStream) {
   MOZ_ASSERT(aCx);
   MOZ_ASSERT(aGlobal);
   MOZ_ASSERT(aStreamReader);
@@ -90,17 +77,24 @@ NS_INTERFACE_MAP_END
     WorkerPrivate* workerPrivate = GetWorkerPrivateFromContext(aCx);
     MOZ_ASSERT(workerPrivate);
 
-    // We need to know when the worker goes away.
-    UniquePtr<FetchStreamReaderWorkerHolder> holder(
-        new FetchStreamReaderWorkerHolder(streamReader));
-    if (NS_WARN_IF(!holder->HoldWorker(workerPrivate, Closing))) {
+    RefPtr<StrongWorkerRef> workerRef = StrongWorkerRef::Create(
+        workerPrivate, "FetchStreamReader", [streamReader]() {
+          MOZ_ASSERT(streamReader);
+          MOZ_ASSERT(streamReader->mWorkerRef);
+
+          streamReader->CloseAndRelease(
+              streamReader->mWorkerRef->Private()->GetJSContext(),
+              NS_ERROR_DOM_INVALID_STATE_ERR);
+        });
+
+    if (NS_WARN_IF(!workerRef)) {
       streamReader->mPipeOut->CloseWithStatus(NS_ERROR_DOM_INVALID_STATE_ERR);
       return NS_ERROR_DOM_INVALID_STATE_ERR;
     }
 
     // These 2 objects create a ref-cycle here that is broken when the stream is
     // closed or the worker shutsdown.
-    streamReader->mWorkerHolder = Move(holder);
+    streamReader->mWorkerRef = std::move(workerRef);
   }
 
   pipeIn.forget(aInputStream);
@@ -115,17 +109,21 @@ FetchStreamReader::FetchStreamReader(nsIGlobalObject* aGlobal)
       mBufferOffset(0),
       mStreamClosed(false) {
   MOZ_ASSERT(aGlobal);
+
+  mozilla::HoldJSObjects(this);
 }
 
 FetchStreamReader::~FetchStreamReader() {
   CloseAndRelease(nullptr, NS_BASE_STREAM_CLOSED);
+
+  mozilla::DropJSObjects(this);
 }
 
 // If a context is provided, an attempt will be made to cancel the reader.  The
 // only situation where we don't expect to have a context is when closure is
-// being triggered from the destructor or the WorkerHolder is notifying.  If
+// being triggered from the destructor or the WorkerRef is notifying.  If
 // we're at the destructor, it's far too late to cancel anything.  And if the
-// WorkerHolder is being notified, the global is going away, so there's also
+// WorkerRef is being notified, the global is going away, so there's also
 // no need to do further JS work.
 void FetchStreamReader::CloseAndRelease(JSContext* aCx, nsresult aStatus) {
   NS_ASSERT_OWNINGTHREAD(FetchStreamReader);
@@ -137,146 +135,166 @@ void FetchStreamReader::CloseAndRelease(JSContext* aCx, nsresult aStatus) {
 
   RefPtr<FetchStreamReader> kungFuDeathGrip = this;
 
-  if (aCx) {
-    MOZ_ASSERT(mReader);
-
+  if (aCx && mReader) {
     RefPtr<DOMException> error = DOMException::Create(aStatus);
 
     JS::Rooted<JS::Value> errorValue(aCx);
     if (ToJSValue(aCx, error, &errorValue)) {
-      JS::Rooted<JSObject*> reader(aCx, mReader);
+      IgnoredErrorResult ignoredError;
       // It's currently safe to cancel an already closed reader because, per the
       // comments in ReadableStream::cancel() conveying the spec, step 2 of
       // 3.4.3 that specified ReadableStreamCancel is: If stream.[[state]] is
       // "closed", return a new promise resolved with undefined.
-      JS::ReadableStreamReaderCancel(aCx, reader, errorValue);
+      RefPtr<Promise> ignoredResultPromise =
+          MOZ_KnownLive(mReader)->Cancel(aCx, errorValue, ignoredError);
+      NS_WARNING_ASSERTION(!ignoredError.Failed(),
+                           "Failed to cancel stream during close and release");
     }
+
+    // We don't want to propagate exceptions during the cleanup.
+    JS_ClearPendingException(aCx);
   }
 
   mStreamClosed = true;
 
   mGlobal = nullptr;
 
-  mPipeOut->CloseWithStatus(aStatus);
+  if (mPipeOut) {
+    mPipeOut->CloseWithStatus(aStatus);
+  }
   mPipeOut = nullptr;
 
-  mWorkerHolder = nullptr;
+  mWorkerRef = nullptr;
 
   mReader = nullptr;
-  mBuffer = nullptr;
+  mBuffer.Clear();
 }
 
-void FetchStreamReader::StartConsuming(JSContext* aCx, JS::HandleObject aStream,
-                                       JS::MutableHandle<JSObject*> aReader,
+void FetchStreamReader::StartConsuming(JSContext* aCx, ReadableStream* aStream,
+                                       ReadableStreamDefaultReader** aReader,
                                        ErrorResult& aRv) {
   MOZ_DIAGNOSTIC_ASSERT(!mReader);
   MOZ_DIAGNOSTIC_ASSERT(aStream);
 
-  JS::Rooted<JSObject*> reader(
-      aCx, JS::ReadableStreamGetReader(aCx, aStream,
-                                       JS::ReadableStreamReaderMode::Default));
-  if (!reader) {
-    aRv.StealExceptionFromJSContext(aCx);
+  RefPtr<ReadableStreamDefaultReader> reader =
+      AcquireReadableStreamDefaultReader(aStream, aRv);
+  if (aRv.Failed()) {
     CloseAndRelease(aCx, NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
 
   mReader = reader;
-  aReader.set(reader);
+  reader.forget(aReader);
 
   aRv = mPipeOut->AsyncWait(this, 0, 0, mOwningEventTarget);
   if (NS_WARN_IF(aRv.Failed())) {
     return;
   }
+  mAsyncWaitWorkerRef = mWorkerRef;
 }
 
-// nsIOutputStreamCallback interface
+struct FetchReadRequest : public ReadRequest {
+ public:
+  NS_DECL_ISUPPORTS_INHERITED
+  NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(FetchReadRequest, ReadRequest)
 
+  explicit FetchReadRequest(FetchStreamReader* aReader)
+      : mFetchStreamReader(aReader) {}
+
+  void ChunkSteps(JSContext* aCx, JS::Handle<JS::Value> aChunk,
+                  ErrorResult& aRv) override {
+    mFetchStreamReader->ChunkSteps(aCx, aChunk, aRv);
+  }
+
+  void CloseSteps(JSContext* aCx, ErrorResult& aRv) override {
+    mFetchStreamReader->CloseAndRelease(aCx, NS_BASE_STREAM_CLOSED);
+  }
+
+  void ErrorSteps(JSContext* aCx, JS::Handle<JS::Value> aError,
+                  ErrorResult& aRv) override {
+    mFetchStreamReader->ErrorSteps(aCx, aError, aRv);
+  }
+
+ protected:
+  virtual ~FetchReadRequest() = default;
+
+  RefPtr<FetchStreamReader> mFetchStreamReader;
+};
+
+NS_IMPL_CYCLE_COLLECTION_INHERITED(FetchReadRequest, ReadRequest,
+                                   mFetchStreamReader)
+NS_IMPL_ADDREF_INHERITED(FetchReadRequest, ReadRequest)
+NS_IMPL_RELEASE_INHERITED(FetchReadRequest, ReadRequest)
+NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(FetchReadRequest)
+NS_INTERFACE_MAP_END_INHERITING(ReadRequest)
+
+// nsIOutputStreamCallback interface
+MOZ_CAN_RUN_SCRIPT_BOUNDARY
 NS_IMETHODIMP
 FetchStreamReader::OnOutputStreamReady(nsIAsyncOutputStream* aStream) {
   NS_ASSERT_OWNINGTHREAD(FetchStreamReader);
-  MOZ_ASSERT(aStream == mPipeOut);
-  MOZ_ASSERT(mReader);
-
+  mAsyncWaitWorkerRef = nullptr;
   if (mStreamClosed) {
     return NS_OK;
   }
 
-  if (mBuffer) {
+  // Only assert if we know the stream is not closed yet.
+  MOZ_ASSERT(aStream == mPipeOut);
+  MOZ_ASSERT(mReader);
+
+  if (!mBuffer.IsEmpty()) {
     return WriteBuffer();
   }
 
-  // TODO: We need to verify this is the correct global per the spec.
-  //       See bug 1385890.
-  AutoEntryScript aes(mGlobal, "ReadableStreamReader.read", !mWorkerHolder);
+  // Here we can retrieve data from the reader using any global we want because
+  // it is not observable. We want to use the reader's global, which is also the
+  // Response's one.
+  AutoEntryScript aes(mGlobal, "ReadableStreamReader.read", !mWorkerRef);
 
-  JS::Rooted<JSObject*> reader(aes.cx(), mReader);
-  JS::Rooted<JSObject*> promise(
-      aes.cx(), JS::ReadableStreamDefaultReaderRead(aes.cx(), reader));
-  if (NS_WARN_IF(!promise)) {
+  IgnoredErrorResult rv;
+
+  // https://fetch.spec.whatwg.org/#incrementally-read-loop
+  // The below very loosely tries to implement the incrementally-read-loop from
+  // the fetch spec.
+  RefPtr<ReadRequest> readRequest = new FetchReadRequest(this);
+  ReadableStreamDefaultReaderRead(aes.cx(), MOZ_KnownLive(mReader), readRequest,
+                                  rv);
+
+  if (NS_WARN_IF(rv.Failed())) {
     // Let's close the stream.
     CloseAndRelease(aes.cx(), NS_ERROR_DOM_INVALID_STATE_ERR);
     return NS_ERROR_FAILURE;
   }
 
-  RefPtr<Promise> domPromise = Promise::CreateFromExisting(mGlobal, promise);
-  if (NS_WARN_IF(!domPromise)) {
-    // Let's close the stream.
-    CloseAndRelease(aes.cx(), NS_ERROR_DOM_INVALID_STATE_ERR);
-    return NS_ERROR_FAILURE;
-  }
-
-  // Let's wait.
-  domPromise->AppendNativeHandler(this);
   return NS_OK;
 }
 
-void FetchStreamReader::ResolvedCallback(JSContext* aCx,
-                                         JS::Handle<JS::Value> aValue) {
-  if (mStreamClosed) {
-    return;
-  }
+void FetchStreamReader::ChunkSteps(JSContext* aCx, JS::Handle<JS::Value> aChunk,
+                                   ErrorResult& aRv) {
+  // This roughly implements the chunk steps from
+  // https://fetch.spec.whatwg.org/#incrementally-read-loop.
 
-  // This promise should be resolved with { done: boolean, value: something },
-  // "value" is interesting only if done is false.
-
-  // We don't want to play with JS api, let's WebIDL bindings doing it for us.
-  // FetchReadableStreamReadDataDone is a dictionary with just a boolean, if the
-  // parsing succeeded, we can proceed with the parsing of the "value", which it
-  // must be a Uint8Array.
-  FetchReadableStreamReadDataDone valueDone;
-  if (!valueDone.Init(aCx, aValue)) {
-    JS_ClearPendingException(aCx);
+  RootedSpiderMonkeyInterface<Uint8Array> chunk(aCx);
+  if (!aChunk.isObject() || !chunk.Init(&aChunk.toObject())) {
     CloseAndRelease(aCx, NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
+  chunk.ComputeState();
 
-  if (valueDone.mDone) {
-    // Stream is completed.
-    CloseAndRelease(aCx, NS_BASE_STREAM_CLOSED);
-    return;
-  }
-
-  UniquePtr<FetchReadableStreamReadDataArray> value(
-      new FetchReadableStreamReadDataArray);
-  if (!value->Init(aCx, aValue) || !value->mValue.WasPassed()) {
-    JS_ClearPendingException(aCx);
-    CloseAndRelease(aCx, NS_ERROR_DOM_INVALID_STATE_ERR);
-    return;
-  }
-
-  Uint8Array& array = value->mValue.Value();
-  array.ComputeLengthAndData();
-  uint32_t len = array.Length();
-
+  uint32_t len = chunk.Length();
   if (len == 0) {
     // If there is nothing to read, let's do another reading.
     OnOutputStreamReady(mPipeOut);
     return;
   }
 
-  MOZ_DIAGNOSTIC_ASSERT(!mBuffer);
-  mBuffer = Move(value);
+  MOZ_DIAGNOSTIC_ASSERT(mBuffer.IsEmpty());
+
+  // Let's take a copy of the data.
+  if (!mBuffer.AppendElements(chunk.Data(), len, fallible)) {
+    CloseAndRelease(aCx, NS_ERROR_OUT_OF_MEMORY);
+    return;
+  }
 
   mBufferOffset = 0;
   mBufferRemaining = len;
@@ -289,12 +307,16 @@ void FetchStreamReader::ResolvedCallback(JSContext* aCx,
   }
 }
 
-nsresult FetchStreamReader::WriteBuffer() {
-  MOZ_ASSERT(mBuffer);
-  MOZ_ASSERT(mBuffer->mValue.WasPassed());
+void FetchStreamReader::ErrorSteps(JSContext* aCx, JS::Handle<JS::Value> aError,
+                                   ErrorResult& aRv) {
+  ReportErrorToConsole(aCx, aError);
+  CloseAndRelease(aCx, NS_ERROR_FAILURE);
+}
 
-  Uint8Array& array = mBuffer->mValue.Value();
-  char* data = reinterpret_cast<char*>(array.Data());
+nsresult FetchStreamReader::WriteBuffer() {
+  MOZ_ASSERT(!mBuffer.IsEmpty());
+
+  char* data = reinterpret_cast<char*>(mBuffer.Elements());
 
   while (1) {
     uint32_t written = 0;
@@ -314,7 +336,7 @@ nsresult FetchStreamReader::WriteBuffer() {
     mBufferOffset += written;
 
     if (mBufferRemaining == 0) {
-      mBuffer = nullptr;
+      mBuffer.Clear();
       break;
     }
   }
@@ -327,12 +349,6 @@ nsresult FetchStreamReader::WriteBuffer() {
   return NS_OK;
 }
 
-void FetchStreamReader::RejectedCallback(JSContext* aCx,
-                                         JS::Handle<JS::Value> aValue) {
-  ReportErrorToConsole(aCx, aValue);
-  CloseAndRelease(aCx, NS_ERROR_FAILURE);
-}
-
 void FetchStreamReader::ReportErrorToConsole(JSContext* aCx,
                                              JS::Handle<JS::Value> aValue) {
   nsCString sourceSpec;
@@ -347,11 +363,10 @@ void FetchStreamReader::ReportErrorToConsole(JSContext* aCx,
   params.AppendElement(valueString);
 
   RefPtr<ConsoleReportCollector> reporter = new ConsoleReportCollector();
-  reporter->AddConsoleReport(
-      nsIScriptError::errorFlag,
-      NS_LITERAL_CSTRING("ReadableStreamReader.read"),
-      nsContentUtils::eDOM_PROPERTIES, sourceSpec, line, column,
-      NS_LITERAL_CSTRING("ReadableStreamReadingFailed"), params);
+  reporter->AddConsoleReport(nsIScriptError::errorFlag,
+                             "ReadableStreamReader.read"_ns,
+                             nsContentUtils::eDOM_PROPERTIES, sourceSpec, line,
+                             column, "ReadableStreamReadingFailed"_ns, params);
 
   uint64_t innerWindowId = 0;
 
@@ -377,5 +392,4 @@ void FetchStreamReader::ReportErrorToConsole(JSContext* aCx,
   workerPrivate->DispatchToMainThread(r.forget());
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
