@@ -4,36 +4,40 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #if !defined(mozilla_dom_HTMLCanvasElement_h)
-#define mozilla_dom_HTMLCanvasElement_h
+#  define mozilla_dom_HTMLCanvasElement_h
 
-#include "mozilla/Attributes.h"
-#include "mozilla/WeakPtr.h"
-#include "nsIDOMEventListener.h"
-#include "nsIObserver.h"
-#include "nsGenericHTMLElement.h"
-#include "nsGkAtoms.h"
-#include "nsSize.h"
-#include "nsError.h"
+#  include "mozilla/Attributes.h"
+#  include "mozilla/StateWatching.h"
+#  include "mozilla/WeakPtr.h"
+#  include "nsIDOMEventListener.h"
+#  include "nsIObserver.h"
+#  include "nsGenericHTMLElement.h"
+#  include "nsGkAtoms.h"
+#  include "nsSize.h"
+#  include "nsError.h"
 
-#include "mozilla/dom/BindingDeclarations.h"
-#include "mozilla/dom/CanvasRenderingContextHelper.h"
-#include "mozilla/gfx/Rect.h"
-#include "mozilla/layers/LayersTypes.h"
+#  include "mozilla/dom/CanvasRenderingContextHelper.h"
+#  include "mozilla/gfx/Rect.h"
+#  include "mozilla/layers/LayersTypes.h"
 
 class nsICanvasRenderingContextInternal;
+class nsIInputStream;
 class nsITimerCallback;
+enum class gfxAlphaType;
+enum class FrameCaptureState : uint8_t;
 
 namespace mozilla {
 
-class WebGLContext;
+class nsDisplayListBuilder;
+class ClientWebGLContext;
 
 namespace layers {
-class AsyncCanvasRenderer;
 class CanvasRenderer;
-class CanvasLayer;
 class Image;
+class ImageContainer;
 class Layer;
 class LayerManager;
+class OOPCanvasRenderer;
 class SharedSurfaceTextureClient;
 class WebRenderCanvasData;
 }  // namespace layers
@@ -41,6 +45,9 @@ namespace gfx {
 class SourceSurface;
 class VRLayerChild;
 }  // namespace gfx
+namespace webgpu {
+class CanvasContext;
+}  // namespace webgpu
 
 namespace dom {
 class BlobCallback;
@@ -48,26 +55,23 @@ class CanvasCaptureMediaStream;
 class File;
 class HTMLCanvasPrintState;
 class OffscreenCanvas;
+class OffscreenCanvasDisplayHelper;
 class PrintCallback;
+class PWebGLChild;
 class RequestedFrameRefreshObserver;
 
 // Listen visibilitychange and memory-pressure event and inform
 // context when event is fired.
-class HTMLCanvasElementObserver final : public nsIObserver,
-                                        public nsIDOMEventListener {
+class HTMLCanvasElementObserver final : public nsIObserver {
  public:
   NS_DECL_ISUPPORTS
   NS_DECL_NSIOBSERVER
-  NS_DECL_NSIDOMEVENTLISTENER
 
   explicit HTMLCanvasElementObserver(HTMLCanvasElement* aElement);
   void Destroy();
 
-  void RegisterVisibilityChangeEvent();
-  void UnregisterVisibilityChangeEvent();
-
-  void RegisterMemoryPressureEvent();
-  void UnregisterMemoryPressureEvent();
+  void RegisterObserverEvents();
+  void UnregisterObserverEvents();
 
  private:
   ~HTMLCanvasElementObserver();
@@ -79,25 +83,18 @@ class HTMLCanvasElementObserver final : public nsIObserver,
  * FrameCaptureListener is used by captureStream() as a way of getting video
  * frames from the canvas. On a refresh driver tick after something has been
  * drawn to the canvas since the last such tick, all registered
- * FrameCaptureListeners whose `mFrameCaptureRequested` equals `true`,
- * will be given a copy of the just-painted canvas.
+ * FrameCaptureListeners that report true for FrameCaptureRequested() will be
+ * given a copy of the just-painted canvas.
  * All FrameCaptureListeners get the same copy.
  */
-class FrameCaptureListener : public SupportsWeakPtr<FrameCaptureListener> {
+class FrameCaptureListener : public SupportsWeakPtr {
  public:
-  MOZ_DECLARE_WEAKREFERENCE_TYPENAME(FrameCaptureListener)
-
-  FrameCaptureListener() : mFrameCaptureRequested(false) {}
-
-  /*
-   * Called when a frame capture is desired on next paint.
-   */
-  void RequestFrameCapture() { mFrameCaptureRequested = true; }
+  FrameCaptureListener() = default;
 
   /*
    * Indicates to the canvas whether or not this listener has requested a frame.
    */
-  bool FrameCaptureRequested() const { return mFrameCaptureRequested; }
+  virtual bool FrameCaptureRequested(const TimeStamp& aTime) const = 0;
 
   /*
    * Interface through which new video frames will be provided while
@@ -107,27 +104,24 @@ class FrameCaptureListener : public SupportsWeakPtr<FrameCaptureListener> {
                         const TimeStamp& aTime) = 0;
 
  protected:
-  virtual ~FrameCaptureListener() {}
-
-  bool mFrameCaptureRequested;
+  virtual ~FrameCaptureListener() = default;
 };
 
 class HTMLCanvasElement final : public nsGenericHTMLElement,
-                                public CanvasRenderingContextHelper {
+                                public CanvasRenderingContextHelper,
+                                public SupportsWeakPtr {
   enum { DEFAULT_CANVAS_WIDTH = 300, DEFAULT_CANVAS_HEIGHT = 150 };
 
-  typedef layers::AsyncCanvasRenderer AsyncCanvasRenderer;
   typedef layers::CanvasRenderer CanvasRenderer;
-  typedef layers::CanvasLayer CanvasLayer;
   typedef layers::Layer Layer;
   typedef layers::LayerManager LayerManager;
   typedef layers::WebRenderCanvasData WebRenderCanvasData;
 
  public:
   explicit HTMLCanvasElement(
-      already_AddRefed<mozilla::dom::NodeInfo>& aNodeInfo);
+      already_AddRefed<mozilla::dom::NodeInfo>&& aNodeInfo);
 
-  NS_IMPL_FROMCONTENT_HTML_WITH_TAG(HTMLCanvasElement, canvas)
+  NS_IMPL_FROMNODE_HTML_WITH_TAG(HTMLCanvasElement, canvas)
 
   // nsISupports
   NS_DECL_ISUPPORTS_INHERITED
@@ -140,36 +134,23 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
   uint32_t Height() {
     return GetUnsignedIntAttr(nsGkAtoms::height, DEFAULT_CANVAS_HEIGHT);
   }
-  void SetHeight(uint32_t aHeight, ErrorResult& aRv) {
-    if (mOffscreenCanvas) {
-      aRv.Throw(NS_ERROR_FAILURE);
-      return;
-    }
-
-    SetUnsignedIntAttr(nsGkAtoms::height, aHeight, DEFAULT_CANVAS_HEIGHT, aRv);
-  }
   uint32_t Width() {
     return GetUnsignedIntAttr(nsGkAtoms::width, DEFAULT_CANVAS_WIDTH);
   }
-  void SetWidth(uint32_t aWidth, ErrorResult& aRv) {
-    if (mOffscreenCanvas) {
-      aRv.Throw(NS_ERROR_FAILURE);
-      return;
-    }
+  void SetHeight(uint32_t aHeight, ErrorResult& aRv);
+  void SetWidth(uint32_t aWidth, ErrorResult& aRv);
 
-    SetUnsignedIntAttr(nsGkAtoms::width, aWidth, DEFAULT_CANVAS_WIDTH, aRv);
-  }
-
-  virtual already_AddRefed<nsISupports> GetContext(
+  already_AddRefed<nsISupports> GetContext(
       JSContext* aCx, const nsAString& aContextId,
-      JS::Handle<JS::Value> aContextOptions, ErrorResult& aRv) override;
+      JS::Handle<JS::Value> aContextOptions, ErrorResult& aRv);
 
   void ToDataURL(JSContext* aCx, const nsAString& aType,
                  JS::Handle<JS::Value> aParams, nsAString& aDataURL,
-                 ErrorResult& aRv);
+                 nsIPrincipal& aSubjectPrincipal, ErrorResult& aRv);
 
   void ToBlob(JSContext* aCx, BlobCallback& aCallback, const nsAString& aType,
-              JS::Handle<JS::Value> aParams, ErrorResult& aRv);
+              JS::Handle<JS::Value> aParams, nsIPrincipal& aSubjectPrincipal,
+              ErrorResult& aRv);
 
   OffscreenCanvas* TransferControlToOffscreen(ErrorResult& aRv);
 
@@ -182,16 +163,14 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
 
     SetHTMLBoolAttr(nsGkAtoms::moz_opaque, aValue, aRv);
   }
-  already_AddRefed<File> MozGetAsFile(const nsAString& aName,
-                                      const nsAString& aType,
-                                      CallerType aCallerType, ErrorResult& aRv);
   already_AddRefed<nsISupports> MozGetIPCContext(const nsAString& aContextId,
                                                  ErrorResult& aRv);
   PrintCallback* GetMozPrintCallback() const;
   void SetMozPrintCallback(PrintCallback* aCallback);
 
   already_AddRefed<CanvasCaptureMediaStream> CaptureStream(
-      const Optional<double>& aFrameRate, ErrorResult& aRv);
+      const Optional<double>& aFrameRate, nsIPrincipal& aSubjectPrincipal,
+      ErrorResult& aRv);
 
   /**
    * Get the size in pixels of this canvas element
@@ -204,15 +183,16 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
   bool IsWriteOnly() const;
 
   /**
-   * Force the canvas to be write-only.
+   * Force the canvas to be write-only, except for readers from
+   * a specific extension's content script expanded principal, if
+   * available.
    */
-  void SetWriteOnly();
+  void SetWriteOnly(nsIPrincipal* aExpandedReader = nullptr);
 
   /**
-   * Force the canvas to be write-only, except for readers from
-   * a specific extension's content script expanded principal.
+   * Notify the placeholder offscreen canvas of an updated size.
    */
-  void SetWriteOnly(nsIPrincipal* aExpandedReader);
+  void InvalidateCanvasPlaceholder(uint32_t aWidth, uint32_t aHeight);
 
   /**
    * Notify that some canvas content has changed and the window may
@@ -225,12 +205,9 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
    */
   void InvalidateCanvas();
 
-  /*
-   * Get the number of contexts in this canvas, and request a context at
-   * an index.
-   */
-  int32_t CountContexts();
-  nsICanvasRenderingContextInternal* GetContextAtIndex(int32_t index);
+  nsICanvasRenderingContextInternal* GetCurrentContext() {
+    return mCurrentContext;
+  }
 
   /*
    * Returns true if the canvas context content is guaranteed to be opaque
@@ -257,7 +234,7 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
    * Returns true when there is at least one registered FrameCaptureListener
    * that has requested a frame capture.
    */
-  bool IsFrameCaptureRequested() const;
+  bool IsFrameCaptureRequested(const TimeStamp& aTime) const;
 
   /*
    * Processes destroyed FrameCaptureListeners and removes them if necessary.
@@ -277,31 +254,28 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
                               const nsAString& aValue,
                               nsIPrincipal* aMaybeScriptedPrincipal,
                               nsAttrValue& aResult) override;
+  NS_IMETHOD_(bool) IsAttributeMapped(const nsAtom* aAttribute) const override;
   nsChangeHint GetAttributeChangeHint(const nsAtom* aAttribute,
                                       int32_t aModType) const override;
+  nsMapRuleToAttributesFunc GetAttributeMappingFunction() const override;
 
-  virtual nsresult Clone(mozilla::dom::NodeInfo* aNodeInfo, nsINode** aResult,
-                         bool aPreallocateChildren) const override;
-  nsresult CopyInnerTo(mozilla::dom::Element* aDest, bool aPreallocateChildren);
+  virtual nsresult Clone(dom::NodeInfo*, nsINode** aResult) const override;
+  nsresult CopyInnerTo(HTMLCanvasElement* aDest);
 
-  virtual nsresult GetEventTargetParent(
-      mozilla::EventChainPreVisitor& aVisitor) override;
+  void GetEventTargetParent(EventChainPreVisitor& aVisitor) override;
+
+  static void MapAttributesIntoRule(const nsMappedAttributes* aAttributes,
+                                    MappedDeclarations&);
 
   /*
    * Helpers called by various users of Canvas
    */
 
-  already_AddRefed<Layer> GetCanvasLayer(nsDisplayListBuilder* aBuilder,
-                                         Layer* aOldLayer,
-                                         LayerManager* aManager);
+  already_AddRefed<layers::Image> GetAsImage();
   bool UpdateWebRenderCanvasData(nsDisplayListBuilder* aBuilder,
                                  WebRenderCanvasData* aCanvasData);
   bool InitializeCanvasRenderer(nsDisplayListBuilder* aBuilder,
                                 CanvasRenderer* aRenderer);
-  // Should return true if the canvas layer should always be marked inactive.
-  // We should return true here if we can't do accelerated compositing with
-  // a non-BasicCanvasLayer.
-  bool ShouldForceInactiveLayer(LayerManager* aManager);
 
   // Call this whenever we need future changes to the canvas
   // to trigger fresh invalidation requests. This needs to be called
@@ -313,24 +287,25 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
   // copies for future frames when no drawing has occurred.
   void MarkContextCleanForFrameCapture();
 
-  // Starts returning false when something is drawn.
-  bool IsContextCleanForFrameCapture();
+  // Returns non-null when the current context supports captureStream().
+  // The FrameCaptureState gets set to DIRTY when something is drawn.
+  Watchable<FrameCaptureState>* GetFrameCaptureState();
 
   nsresult GetContext(const nsAString& aContextId, nsISupports** aContext);
 
   layers::LayersBackend GetCompositorBackendType() const;
 
-  void OnVisibilityChange();
-
   void OnMemoryPressure();
-
-  static void SetAttrFromAsyncCanvasRenderer(AsyncCanvasRenderer* aRenderer);
-  static void InvalidateFromAsyncCanvasRenderer(AsyncCanvasRenderer* aRenderer);
+  void OnDeviceReset();
 
   already_AddRefed<layers::SharedSurfaceTextureClient> GetVRFrame();
+  void ClearVRFrame();
+
+  bool MaybeModified() const { return mMaybeModified; };
 
  protected:
   virtual ~HTMLCanvasElement();
+  void Destroy();
 
   virtual JSObject* WrapNode(JSContext* aCx,
                              JS::Handle<JSObject*> aGivenProto) override;
@@ -340,13 +315,13 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
   virtual already_AddRefed<nsICanvasRenderingContextInternal> CreateContext(
       CanvasContextType aContextType) override;
 
-  nsresult ExtractData(JSContext* aCx, nsAString& aType,
-                       const nsAString& aOptions, nsIInputStream** aStream);
-  nsresult ToDataURLImpl(JSContext* aCx, const nsAString& aMimeType,
+  nsresult ExtractData(JSContext* aCx, nsIPrincipal& aSubjectPrincipal,
+                       nsAString& aType, const nsAString& aOptions,
+                       nsIInputStream** aStream);
+  nsresult ToDataURLImpl(JSContext* aCx, nsIPrincipal& aSubjectPrincipal,
+                         const nsAString& aMimeType,
                          const JS::Value& aEncoderOptions, nsAString& aDataURL);
-  nsresult MozGetAsFileImpl(const nsAString& aName, const nsAString& aType,
-                            File** aResult);
-  void CallPrintCallback();
+  MOZ_CAN_RUN_SCRIPT void CallPrintCallback();
 
   virtual nsresult AfterSetAttr(int32_t aNamespaceID, nsAtom* aName,
                                 const nsAttrValue* aValue,
@@ -357,16 +332,30 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
                                           const nsAttrValueOrString& aValue,
                                           bool aNotify) override;
 
-  AsyncCanvasRenderer* GetAsyncCanvasRenderer();
+ public:
+  ClientWebGLContext* GetWebGLContext();
+  webgpu::CanvasContext* GetWebGPUContext();
 
+  bool IsOffscreen() const { return !!mOffscreenCanvas; }
+  OffscreenCanvas* GetOffscreenCanvas() const { return mOffscreenCanvas; }
+
+  layers::ImageContainer* GetImageContainer() const { return mImageContainer; }
+
+  layers::CompositableHandle GetCompositableHandle() const;
+
+ protected:
   bool mResetLayer;
+  bool mMaybeModified;  // we fetched the context, so we may have written to the
+                        // canvas
   RefPtr<HTMLCanvasElement> mOriginalCanvas;
   RefPtr<PrintCallback> mPrintCallback;
   RefPtr<HTMLCanvasPrintState> mPrintState;
   nsTArray<WeakPtr<FrameCaptureListener>> mRequestedFrameListeners;
   RefPtr<RequestedFrameRefreshObserver> mRequestedFrameRefreshObserver;
-  RefPtr<AsyncCanvasRenderer> mAsyncCanvasRenderer;
+  RefPtr<CanvasRenderer> mCanvasRenderer;
   RefPtr<OffscreenCanvas> mOffscreenCanvas;
+  RefPtr<OffscreenCanvasDisplayHelper> mOffscreenDisplay;
+  RefPtr<layers::ImageContainer> mImageContainer;
   RefPtr<HTMLCanvasElementObserver> mContextObserver;
 
  public:
@@ -375,17 +364,17 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
   // We also transitively set it when script paints a canvas which
   // is itself write-only.
   bool mWriteOnly;
- 
+
   // When this canvas is (only) tainted by an image from an extension
   // content script, allow reads from the same extension afterwards.
   RefPtr<nsIPrincipal> mExpandedReader;
 
   // Determines if the caller should be able to read the content.
-  bool CallerCanRead(JSContext* aCx);
-  
+  bool CallerCanRead(JSContext* aCx) const;
+
   bool IsPrintCallbackDone();
 
-  void HandlePrintCallback(nsPresContext::nsPresContextType aType);
+  void HandlePrintCallback(nsPresContext*);
 
   nsresult DispatchPrintCallback(nsITimerCallback* aCallback);
 
@@ -393,7 +382,7 @@ class HTMLCanvasElement final : public nsGenericHTMLElement,
 
   HTMLCanvasElement* GetOriginalCanvas();
 
-  CanvasContextType GetCurrentContextType() { return mCurrentContextType; }
+  CanvasContextType GetCurrentContextType();
 
  private:
   /**
@@ -423,7 +412,7 @@ class HTMLCanvasPrintState final : public nsWrapperCache {
   bool mIsDone;
 
   NS_INLINE_DECL_CYCLE_COLLECTING_NATIVE_REFCOUNTING(HTMLCanvasPrintState)
-  NS_DECL_CYCLE_COLLECTION_SCRIPT_HOLDER_NATIVE_CLASS(HTMLCanvasPrintState)
+  NS_DECL_CYCLE_COLLECTION_NATIVE_WRAPPERCACHE_CLASS(HTMLCanvasPrintState)
 
   virtual JSObject* WrapObject(JSContext* cx,
                                JS::Handle<JSObject*> aGivenProto) override;
