@@ -10,21 +10,18 @@
 #include "mozilla/Attributes.h"
 #include "nsGenericHTMLElement.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 class HTMLMetaElement final : public nsGenericHTMLElement {
  public:
-  explicit HTMLMetaElement(already_AddRefed<mozilla::dom::NodeInfo>& aNodeInfo);
+  explicit HTMLMetaElement(
+      already_AddRefed<mozilla::dom::NodeInfo>&& aNodeInfo);
 
   // nsISupports
   NS_INLINE_DECL_REFCOUNTING_INHERITED(HTMLMetaElement, nsGenericHTMLElement)
 
-  virtual nsresult BindToTree(nsIDocument* aDocument, nsIContent* aParent,
-                              nsIContent* aBindingParent,
-                              bool aCompileEventHandlers) override;
-  virtual void UnbindFromTree(bool aDeep = true,
-                              bool aNullParent = true) override;
+  virtual nsresult BindToTree(BindContext&, nsINode& aParent) override;
+  virtual void UnbindFromTree(bool aNullParent = true) override;
 
   virtual nsresult AfterSetAttr(int32_t aNameSpaceID, nsAtom* aName,
                                 const nsAttrValue* aValue,
@@ -32,10 +29,9 @@ class HTMLMetaElement final : public nsGenericHTMLElement {
                                 nsIPrincipal* aSubjectPrincipal,
                                 bool aNotify) override;
 
-  void CreateAndDispatchEvent(nsIDocument* aDoc, const nsAString& aEventName);
+  void CreateAndDispatchEvent(Document&, const nsAString& aEventName);
 
-  virtual nsresult Clone(mozilla::dom::NodeInfo* aNodeInfo, nsINode** aResult,
-                         bool aPreallocateChildren) const override;
+  virtual nsresult Clone(dom::NodeInfo*, nsINode** aResult) const override;
 
   void GetName(nsAString& aValue) { GetHTMLAttr(nsGkAtoms::name, aValue); }
   void SetName(const nsAString& aName, ErrorResult& aRv) {
@@ -47,9 +43,8 @@ class HTMLMetaElement final : public nsGenericHTMLElement {
   void SetHttpEquiv(const nsAString& aHttpEquiv, ErrorResult& aRv) {
     SetHTMLAttr(nsGkAtoms::httpEquiv, aHttpEquiv, aRv);
   }
-  nsresult GetContent(nsAString& aValue) {
+  void GetContent(nsAString& aValue) {
     GetHTMLAttr(nsGkAtoms::content, aValue);
-    return NS_OK;
   }
   void SetContent(const nsAString& aContent, ErrorResult& aRv) {
     SetHTMLAttr(nsGkAtoms::content, aContent, aRv);
@@ -66,10 +61,13 @@ class HTMLMetaElement final : public nsGenericHTMLElement {
   virtual ~HTMLMetaElement();
 
  private:
-  nsresult SetMetaReferrer(nsIDocument* aDocument);
+  enum class ChangeKind : uint8_t { TreeChange, NameChange, ContentChange };
+  void MetaRemoved(Document& aDoc, const nsAttrValue& aName,
+                   ChangeKind aChangeKind);
+  void MetaAddedOrChanged(Document& aDoc, const nsAttrValue& aName,
+                          ChangeKind aChangeKind);
 };
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
 
 #endif  // mozilla_dom_HTMLMetaElement_h
