@@ -10,8 +10,7 @@
 
 NS_IMPL_NS_NEW_HTML_ELEMENT(Progress)
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 const double HTMLProgressElement::kIndeterminatePosition = -1.0;
 const double HTMLProgressElement::kDefaultValue = 0.0;
@@ -31,7 +30,8 @@ NS_IMPL_ELEMENT_CLONE(HTMLProgressElement)
 EventStates HTMLProgressElement::IntrinsicState() const {
   EventStates state = nsGenericHTMLElement::IntrinsicState();
 
-  if (IsIndeterminate()) {
+  const nsAttrValue* attrValue = mAttrs.GetAttr(nsGkAtoms::value);
+  if (!attrValue || attrValue->Type() != nsAttrValue::eDoubleValue) {
     state |= NS_EVENT_STATE_INDETERMINATE;
   }
 
@@ -74,22 +74,16 @@ double HTMLProgressElement::Max() const {
 }
 
 double HTMLProgressElement::Position() const {
-  if (IsIndeterminate()) {
+  if (State().HasState(NS_EVENT_STATE_INDETERMINATE)) {
     return kIndeterminatePosition;
   }
 
   return Value() / Max();
 }
 
-bool HTMLProgressElement::IsIndeterminate() const {
-  const nsAttrValue* attrValue = mAttrs.GetAttr(nsGkAtoms::value);
-  return !attrValue || attrValue->Type() != nsAttrValue::eDoubleValue;
-}
-
 JSObject* HTMLProgressElement::WrapNode(JSContext* aCx,
                                         JS::Handle<JSObject*> aGivenProto) {
   return HTMLProgressElement_Binding::Wrap(aCx, this, aGivenProto);
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
