@@ -10,10 +10,10 @@
 #include "mozilla/AsyncEventDispatcher.h"
 #include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/dom/HTMLInputElement.h"
+#include "mozilla/dom/ShadowRoot.h"
 #include "nsDOMTokenList.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 const double DateTimeInputTypeBase::kMinimumYear = 1;
 const double DateTimeInputTypeBase::kMaximumYear = 275760;
@@ -23,7 +23,7 @@ const double DateTimeInputTypeBase::kMsPerDay = 24 * 60 * 60 * 1000;
 
 bool DateTimeInputTypeBase::IsMutable() const {
   return !mInputElement->IsDisabled() &&
-         !mInputElement->HasAttr(kNameSpaceID_None, nsGkAtoms::readonly);
+         !mInputElement->HasAttr(nsGkAtoms::readonly);
 }
 
 bool DateTimeInputTypeBase::IsValueMissing() const {
@@ -88,41 +88,39 @@ bool DateTimeInputTypeBase::HasStepMismatch(bool aUseZeroIfValueNaN) const {
 }
 
 bool DateTimeInputTypeBase::HasBadInput() const {
-  if (!mInputElement->GetShadowRoot()) {
+  ShadowRoot* shadow = mInputElement->GetShadowRoot();
+  if (!shadow) {
     return false;
   }
 
-  Element* editWrapperElement = mInputElement->GetShadowRoot()->GetElementById(
-      NS_LITERAL_STRING("edit-wrapper"));
-
+  Element* editWrapperElement = shadow->GetElementById(u"edit-wrapper"_ns);
   if (!editWrapperElement) {
     return false;
   }
 
-  // Incomplete field does not imply bad input.
+  bool allEmpty = true;
+  // Empty field does not imply bad input, but incomplete field does.
   for (Element* child = editWrapperElement->GetFirstElementChild(); child;
        child = child->GetNextElementSibling()) {
-    if (child->ClassList()->Contains(
-            NS_LITERAL_STRING("datetime-edit-field"))) {
-      nsAutoString value;
-      child->GetAttr(kNameSpaceID_None, nsGkAtoms::value, value);
-      if (value.IsEmpty()) {
-        return false;
-      }
+    if (!child->ClassList()->Contains(u"datetime-edit-field"_ns)) {
+      continue;
+    }
+    nsAutoString value;
+    child->GetAttr(nsGkAtoms::value, value);
+    if (!value.IsEmpty()) {
+      allEmpty = false;
+      break;
     }
   }
 
-  // All fields are available but input element's value is empty implies
-  // it has been sanitized.
-  nsAutoString value;
-  mInputElement->GetValue(value, CallerType::System);
-
-  return value.IsEmpty();
+  // If some fields are available but input element's value is empty implies it
+  // has been sanitized.
+  return !allEmpty && IsValueEmpty();
 }
 
 nsresult DateTimeInputTypeBase::GetRangeOverflowMessage(nsAString& aMessage) {
   nsAutoString maxStr;
-  mInputElement->GetAttr(kNameSpaceID_None, nsGkAtoms::max, maxStr);
+  mInputElement->GetAttr(nsGkAtoms::max, maxStr);
 
   return nsContentUtils::FormatMaybeLocalizedString(
       aMessage, nsContentUtils::eDOM_PROPERTIES,
@@ -131,7 +129,7 @@ nsresult DateTimeInputTypeBase::GetRangeOverflowMessage(nsAString& aMessage) {
 
 nsresult DateTimeInputTypeBase::GetRangeUnderflowMessage(nsAString& aMessage) {
   nsAutoString minStr;
-  mInputElement->GetAttr(kNameSpaceID_None, nsGkAtoms::min, minStr);
+  mInputElement->GetAttr(nsGkAtoms::min, minStr);
 
   return nsContentUtils::FormatMaybeLocalizedString(
       aMessage, nsContentUtils::eDOM_PROPERTIES,
@@ -142,7 +140,7 @@ nsresult DateTimeInputTypeBase::GetRangeUnderflowMessage(nsAString& aMessage) {
 nsresult DateTimeInputTypeBase::MinMaxStepAttrChanged() {
   if (Element* dateTimeBoxElement = mInputElement->GetDateTimeBoxElement()) {
     AsyncEventDispatcher* dispatcher = new AsyncEventDispatcher(
-        dateTimeBoxElement, NS_LITERAL_STRING("MozNotifyMinMaxStepAttrChanged"),
+        dateTimeBoxElement, u"MozNotifyMinMaxStepAttrChanged"_ns,
         CanBubble::eNo, ChromeOnlyDispatch::eNo);
     dispatcher->RunDOMEventWhenSafe();
   }
@@ -176,10 +174,6 @@ bool DateTimeInputTypeBase::GetTimeFromMs(double aValue, uint16_t* aHours,
 // input type=date
 
 nsresult DateInputType::GetBadInputMessage(nsAString& aMessage) {
-  if (!StaticPrefs::dom_forms_datetime()) {
-    return NS_ERROR_UNEXPECTED;
-  }
-
   return nsContentUtils::GetMaybeLocalizedString(
       nsContentUtils::eDOM_PROPERTIES, "FormValidationInvalidDate",
       mInputElement->OwnerDoc(), aMessage);
@@ -224,6 +218,12 @@ bool DateInputType::ConvertNumberToString(Decimal aValue,
 
 // input type=time
 
+nsresult TimeInputType::GetBadInputMessage(nsAString& aMessage) {
+  return nsContentUtils::GetMaybeLocalizedString(
+      nsContentUtils::eDOM_PROPERTIES, "FormValidationInvalidTime",
+      mInputElement->OwnerDoc(), aMessage);
+}
+
 bool TimeInputType::ConvertStringToNumber(nsAString& aValue,
                                           Decimal& aResultValue) const {
   uint32_t milliseconds;
@@ -309,10 +309,10 @@ bool TimeInputType::IsRangeUnderflow() const {
 nsresult TimeInputType::GetReversedRangeUnderflowAndOverflowMessage(
     nsAString& aMessage) {
   nsAutoString maxStr;
-  mInputElement->GetAttr(kNameSpaceID_None, nsGkAtoms::max, maxStr);
+  mInputElement->GetAttr(nsGkAtoms::max, maxStr);
 
   nsAutoString minStr;
-  mInputElement->GetAttr(kNameSpaceID_None, nsGkAtoms::min, minStr);
+  mInputElement->GetAttr(nsGkAtoms::min, minStr);
 
   return nsContentUtils::FormatMaybeLocalizedString(
       aMessage, nsContentUtils::eDOM_PROPERTIES,
@@ -334,6 +334,12 @@ nsresult TimeInputType::GetRangeUnderflowMessage(nsAString& aMessage) {
 
 // input type=week
 
+nsresult WeekInputType::GetBadInputMessage(nsAString& aMessage) {
+  return nsContentUtils::GetMaybeLocalizedString(
+      nsContentUtils::eDOM_PROPERTIES, "FormValidationInvalidWeek",
+      mInputElement->OwnerDoc(), aMessage);
+}
+
 bool WeekInputType::ConvertStringToNumber(nsAString& aValue,
                                           Decimal& aResultValue) const {
   uint32_t year, week;
@@ -397,6 +403,12 @@ bool WeekInputType::ConvertNumberToString(Decimal aValue,
 
 // input type=month
 
+nsresult MonthInputType::GetBadInputMessage(nsAString& aMessage) {
+  return nsContentUtils::GetMaybeLocalizedString(
+      nsContentUtils::eDOM_PROPERTIES, "FormValidationInvalidMonth",
+      mInputElement->OwnerDoc(), aMessage);
+}
+
 bool MonthInputType::ConvertStringToNumber(nsAString& aValue,
                                            Decimal& aResultValue) const {
   uint32_t year, month;
@@ -446,6 +458,12 @@ bool MonthInputType::ConvertNumberToString(Decimal aValue,
 
 // input type=datetime-local
 
+nsresult DateTimeLocalInputType::GetBadInputMessage(nsAString& aMessage) {
+  return nsContentUtils::GetMaybeLocalizedString(
+      nsContentUtils::eDOM_PROPERTIES, "FormValidationInvalidDateTime",
+      mInputElement->OwnerDoc(), aMessage);
+}
+
 bool DateTimeLocalInputType::ConvertStringToNumber(
     nsAString& aValue, Decimal& aResultValue) const {
   uint32_t year, month, day, timeInMs;
@@ -502,5 +520,4 @@ bool DateTimeLocalInputType::ConvertNumberToString(
   return true;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
