@@ -12,30 +12,24 @@
 #include "StorageUtils.h"
 #include "LocalStorageManager.h"
 
-#include "nsAutoPtr.h"
 #include "nsDOMString.h"
 #include "nsXULAppAPI.h"
 #include "mozilla/Unused.h"
 #include "nsProxyRelease.h"
 #include "nsThreadUtils.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 #define DOM_STORAGE_CACHE_KEEP_ALIVE_TIME_MS 20000
 
 namespace {
 
 const uint32_t kDefaultSet = 0;
-const uint32_t kPrivateSet = 1;
-const uint32_t kSessionSet = 2;
+const uint32_t kSessionSet = 1;
 
-inline uint32_t GetDataSetIndex(bool aPrivate, bool aSessionOnly) {
-  if (aPrivate) {
-    return kPrivateSet;
-  }
-
-  if (aSessionOnly) {
+inline uint32_t GetDataSetIndex(bool aPrivateBrowsing,
+                                bool aSessionScopedOrLess) {
+  if (!aPrivateBrowsing && aSessionScopedOrLess) {
     return kSessionSet;
   }
 
@@ -43,7 +37,8 @@ inline uint32_t GetDataSetIndex(bool aPrivate, bool aSessionOnly) {
 }
 
 inline uint32_t GetDataSetIndex(const LocalStorage* aStorage) {
-  return GetDataSetIndex(aStorage->IsPrivate(), aStorage->IsSessionOnly());
+  return GetDataSetIndex(aStorage->IsPrivateBrowsing(),
+                         aStorage->IsSessionScopedOrLess());
 }
 
 }  // namespace
@@ -77,7 +72,6 @@ LocalStorageCache::LocalStorageCache(const nsACString* aOriginNoSuffix)
       mLoadResult(NS_OK),
       mInitialized(false),
       mPersistent(false),
-      mSessionOnlyDataSetActive(false),
       mPreloadTelemetryRecorded(false) {
   MOZ_COUNT_CTOR(LocalStorageCache);
 }
@@ -128,18 +122,17 @@ LocalStorageCache::Release(void) {
 void LocalStorageCache::Init(LocalStorageManager* aManager, bool aPersistent,
                              nsIPrincipal* aPrincipal,
                              const nsACString& aQuotaOriginScope) {
+  MOZ_ASSERT(!aQuotaOriginScope.IsEmpty());
+
   if (mInitialized) {
     return;
   }
 
   mInitialized = true;
   aPrincipal->OriginAttributesRef().CreateSuffix(mOriginSuffix);
+  mPrivateBrowsingId = aPrincipal->GetPrivateBrowsingId();
   mPersistent = aPersistent;
-  if (aQuotaOriginScope.IsEmpty()) {
-    mQuotaOriginScope = Origin();
-  } else {
-    mQuotaOriginScope = aQuotaOriginScope;
-  }
+  mQuotaOriginScope = aQuotaOriginScope;
 
   if (mPersistent) {
     mManager = aManager;
@@ -150,9 +143,9 @@ void LocalStorageCache::Init(LocalStorageManager* aManager, bool aPersistent,
   // this storage cache is bound to.
   MOZ_ASSERT(StringBeginsWith(mQuotaOriginScope, mOriginSuffix));
   MOZ_ASSERT(mOriginSuffix.IsEmpty() !=
-             StringBeginsWith(mQuotaOriginScope, NS_LITERAL_CSTRING("^")));
+             StringBeginsWith(mQuotaOriginScope, "^"_ns));
 
-  mUsage = aManager->GetOriginUsage(mQuotaOriginScope);
+  mUsage = aManager->GetOriginUsage(mQuotaOriginScope, mPrivateBrowsingId);
 }
 
 void LocalStorageCache::NotifyObservers(const LocalStorage* aStorage,
@@ -174,7 +167,8 @@ void LocalStorageCache::NotifyObservers(const LocalStorage* aStorage,
 }
 
 inline bool LocalStorageCache::Persist(const LocalStorage* aStorage) const {
-  return mPersistent && !aStorage->IsSessionOnly() && !aStorage->IsPrivate();
+  return mPersistent &&
+         (aStorage->IsPrivateBrowsing() || !aStorage->IsSessionScopedOrLess());
 }
 
 const nsCString LocalStorageCache::Origin() const {
@@ -183,29 +177,7 @@ const nsCString LocalStorageCache::Origin() const {
 
 LocalStorageCache::Data& LocalStorageCache::DataSet(
     const LocalStorage* aStorage) {
-  uint32_t index = GetDataSetIndex(aStorage);
-
-  if (index == kSessionSet && !mSessionOnlyDataSetActive) {
-    // Session only data set is demanded but not filled with
-    // current data set, copy to session only set now.
-
-    WaitForPreload(Telemetry::LOCALDOMSTORAGE_SESSIONONLY_PRELOAD_BLOCKING_MS);
-
-    Data& defaultSet = mData[kDefaultSet];
-    Data& sessionSet = mData[kSessionSet];
-
-    for (auto iter = defaultSet.mKeys.Iter(); !iter.Done(); iter.Next()) {
-      sessionSet.mKeys.Put(iter.Key(), iter.UserData());
-    }
-
-    mSessionOnlyDataSetActive = true;
-
-    // This updates sessionSet.mOriginQuotaUsage and also updates global usage
-    // for all session only data
-    ProcessUsageDelta(kSessionSet, defaultSet.mOriginQuotaUsage);
-  }
-
-  return mData[index];
+  return mData[GetDataSetIndex(aStorage)];
 }
 
 bool LocalStorageCache::ProcessUsageDelta(const LocalStorage* aStorage,
@@ -217,17 +189,11 @@ bool LocalStorageCache::ProcessUsageDelta(const LocalStorage* aStorage,
 bool LocalStorageCache::ProcessUsageDelta(uint32_t aGetDataSetIndex,
                                           const int64_t aDelta,
                                           const MutationSource aSource) {
-  // Check if we are in a low disk space situation
-  if (aSource == ContentMutation && aDelta > 0 && mManager &&
-      mManager->IsLowDiskSpace()) {
-    return false;
-  }
-
   // Check limit per this origin
   Data& data = mData[aGetDataSetIndex];
   uint64_t newOriginUsage = data.mOriginQuotaUsage + aDelta;
   if (aSource == ContentMutation && aDelta > 0 &&
-      newOriginUsage > LocalStorageManager::GetQuota()) {
+      newOriginUsage > LocalStorageManager::GetOriginQuota()) {
     return false;
   }
 
@@ -247,7 +213,8 @@ void LocalStorageCache::Preload() {
     return;
   }
 
-  StorageDBChild* storageChild = StorageDBChild::GetOrCreate();
+  StorageDBChild* storageChild =
+      StorageDBChild::GetOrCreate(mPrivateBrowsingId);
   if (!storageChild) {
     mLoaded = true;
     mLoadResult = NS_ERROR_FAILURE;
@@ -257,25 +224,6 @@ void LocalStorageCache::Preload() {
   storageChild->AsyncPreload(this);
 }
 
-namespace {
-
-// The AutoTimer provided by telemetry headers is only using static,
-// i.e. compile time known ID, but here we know the ID only at run time.
-// Hence a new class.
-class TelemetryAutoTimer {
- public:
-  explicit TelemetryAutoTimer(Telemetry::HistogramID aId)
-      : id(aId), start(TimeStamp::Now()) {}
-
-  ~TelemetryAutoTimer() { Telemetry::AccumulateTimeDelta(id, start); }
-
- private:
-  Telemetry::HistogramID id;
-  const TimeStamp start;
-};
-
-}  // namespace
-
 void LocalStorageCache::WaitForPreload(Telemetry::HistogramID aTelemetryID) {
   if (!mPersistent) {
     return;
@@ -295,7 +243,7 @@ void LocalStorageCache::WaitForPreload(Telemetry::HistogramID aTelemetryID) {
   }
 
   // Measure which operation blocks and for how long
-  TelemetryAutoTimer timer(aTelemetryID);
+  Telemetry::RuntimeAutoTimer timer(aTelemetryID);
 
   // If preload already started (i.e. we got some first data, but not all)
   // SyncPreload will just wait for it to finish rather then synchronously
@@ -306,7 +254,7 @@ void LocalStorageCache::WaitForPreload(Telemetry::HistogramID aTelemetryID) {
   // No need to check sDatabase for being non-null since preload is either
   // done before we've shut the DB down or when the DB could not start,
   // preload has not even be started.
-  StorageDBChild::Get()->SyncPreload(this);
+  StorageDBChild::Get(mPrivateBrowsingId)->SyncPreload(this);
 }
 
 nsresult LocalStorageCache::GetLength(const LocalStorage* aStorage,
@@ -357,9 +305,7 @@ void LocalStorageCache::GetKeys(const LocalStorage* aStorage,
     return;
   }
 
-  for (auto iter = DataSet(aStorage).mKeys.Iter(); !iter.Done(); iter.Next()) {
-    aKeys.AppendElement(iter.Key());
-  }
+  AppendToArray(aKeys, DataSet(aStorage).mKeys.Keys());
 }
 
 nsresult LocalStorageCache::GetItem(const LocalStorage* aStorage,
@@ -408,23 +354,25 @@ nsresult LocalStorageCache::SetItem(const LocalStorage* aStorage,
            static_cast<int64_t>(aOld.Length());
 
   if (!ProcessUsageDelta(aStorage, delta, aSource)) {
-    return NS_ERROR_DOM_QUOTA_REACHED;
+    return NS_ERROR_DOM_QUOTA_EXCEEDED_ERR;
   }
 
   if (aValue == aOld && DOMStringIsNull(aValue) == DOMStringIsNull(aOld)) {
     return NS_SUCCESS_DOM_NO_OPERATION;
   }
 
-  data.mKeys.Put(aKey, aValue);
+  data.mKeys.InsertOrUpdate(aKey, aValue);
 
   if (aSource != ContentMutation) {
     return NS_OK;
   }
 
+#if !defined(MOZ_WIDGET_ANDROID)
   NotifyObservers(aStorage, nsString(aKey), aOld, aValue);
+#endif
 
   if (Persist(aStorage)) {
-    StorageDBChild* storageChild = StorageDBChild::Get();
+    StorageDBChild* storageChild = StorageDBChild::Get(mPrivateBrowsingId);
     if (!storageChild) {
       NS_ERROR(
           "Writing to localStorage after the database has been shut down"
@@ -468,10 +416,12 @@ nsresult LocalStorageCache::RemoveItem(const LocalStorage* aStorage,
     return NS_OK;
   }
 
+#if !defined(MOZ_WIDGET_ANDROID)
   NotifyObservers(aStorage, nsString(aKey), aOld, VoidString());
+#endif
 
   if (Persist(aStorage)) {
-    StorageDBChild* storageChild = StorageDBChild::Get();
+    StorageDBChild* storageChild = StorageDBChild::Get(mPrivateBrowsingId);
     if (!storageChild) {
       NS_ERROR(
           "Writing to localStorage after the database has been shut down"
@@ -515,12 +465,14 @@ nsresult LocalStorageCache::Clear(const LocalStorage* aStorage,
     return hadData ? NS_OK : NS_SUCCESS_DOM_NO_OPERATION;
   }
 
+#if !defined(MOZ_WIDGET_ANDROID)
   if (hadData) {
     NotifyObservers(aStorage, VoidString(), VoidString(), VoidString());
   }
+#endif
 
   if (Persist(aStorage) && (refresh || hadData)) {
-    StorageDBChild* storageChild = StorageDBChild::Get();
+    StorageDBChild* storageChild = StorageDBChild::Get(mPrivateBrowsingId);
     if (!storageChild) {
       NS_ERROR(
           "Writing to localStorage after the database has been shut down"
@@ -552,15 +504,9 @@ void LocalStorageCache::UnloadItems(uint32_t aUnloadFlags) {
     ProcessUsageDelta(kDefaultSet, -mData[kDefaultSet].mOriginQuotaUsage);
   }
 
-  if (aUnloadFlags & kUnloadPrivate) {
-    mData[kPrivateSet].mKeys.Clear();
-    ProcessUsageDelta(kPrivateSet, -mData[kPrivateSet].mOriginQuotaUsage);
-  }
-
   if (aUnloadFlags & kUnloadSession) {
     mData[kSessionSet].mKeys.Clear();
     ProcessUsageDelta(kSessionSet, -mData[kSessionSet].mOriginQuotaUsage);
-    mSessionOnlyDataSetActive = false;
   }
 
 #ifdef DOM_STORAGE_TESTS
@@ -590,12 +536,10 @@ bool LocalStorageCache::LoadItem(const nsAString& aKey,
   }
 
   Data& data = mData[kDefaultSet];
-  if (data.mKeys.Get(aKey, nullptr)) {
-    return true;  // don't stop, just don't override
-  }
-
-  data.mKeys.Put(aKey, aValue);
-  data.mOriginQuotaUsage += aKey.Length() + aValue.Length();
+  data.mKeys.LookupOrInsertWith(aKey, [&] {
+    data.mOriginQuotaUsage += aKey.Length() + aValue.Length();
+    return aValue;
+  });
   return true;
 }
 
@@ -617,7 +561,7 @@ void LocalStorageCache::LoadWait() {
 
 StorageUsage::StorageUsage(const nsACString& aOriginScope)
     : mOriginScope(aOriginScope) {
-  mUsage[kDefaultSet] = mUsage[kPrivateSet] = mUsage[kSessionSet] = 0LL;
+  mUsage[kDefaultSet] = mUsage[kSessionSet] = 0LL;
 }
 
 namespace {
@@ -660,7 +604,7 @@ bool StorageUsage::CheckAndSetETLD1UsageDelta(
 
   int64_t newUsage = mUsage[aDataSetIndex] + aDelta;
   if (aSource == LocalStorageCache::ContentMutation && aDelta > 0 &&
-      newUsage > LocalStorageManager::GetQuota()) {
+      newUsage > LocalStorageManager::GetSiteQuota()) {
     return false;
   }
 
@@ -668,5 +612,4 @@ bool StorageUsage::CheckAndSetETLD1UsageDelta(
   return true;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
