@@ -8,8 +8,7 @@
 #include "nsISupportsImpl.h"
 #include "mozilla/dom/DOMString.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 class XMLHttpRequestStringBuffer final {
   friend class XMLHttpRequestStringWriterHelper;
@@ -26,11 +25,13 @@ class XMLHttpRequestStringBuffer final {
     return mData.Length();
   }
 
-  uint32_t UnsafeLength() const { return mData.Length(); }
+  uint32_t UnsafeLength() const NO_THREAD_SAFETY_ANALYSIS {
+    return mData.Length();
+  }
 
-  mozilla::BulkWriteHandle<char16_t> UnsafeBulkWrite(uint32_t aCapacity,
-                                                     nsresult& aRv) {
-    return mData.BulkWrite(aCapacity, UnsafeLength(), false, aRv);
+  mozilla::Result<mozilla::BulkWriteHandle<char16_t>, nsresult> UnsafeBulkWrite(
+      uint32_t aCapacity) NO_THREAD_SAFETY_ANALYSIS {
+    return mData.BulkWrite(aCapacity, UnsafeLength(), false);
   }
 
   void Append(const nsAString& aString) {
@@ -40,16 +41,17 @@ class XMLHttpRequestStringBuffer final {
     mData.Append(aString);
   }
 
-  MOZ_MUST_USE bool GetAsString(nsAString& aString) {
+  [[nodiscard]] bool GetAsString(nsAString& aString) {
     MutexAutoLock lock(mMutex);
     return aString.Assign(mData, mozilla::fallible);
   }
 
-  size_t SizeOfThis(MallocSizeOf aMallocSizeOf) const {
+  size_t SizeOfThis(MallocSizeOf aMallocSizeOf) {
+    MutexAutoLock lock(mMutex);
     return mData.SizeOfExcludingThisIfUnshared(aMallocSizeOf);
   }
 
-  MOZ_MUST_USE bool GetAsString(DOMString& aString, uint32_t aLength) {
+  [[nodiscard]] bool GetAsString(DOMString& aString, uint32_t aLength) {
     MutexAutoLock lock(mMutex);
     MOZ_ASSERT(aLength <= mData.Length());
 
@@ -84,7 +86,7 @@ class XMLHttpRequestStringBuffer final {
   Mutex mMutex;
 
   // The following member variable is protected by mutex.
-  nsString mData;
+  nsString mData GUARDED_BY(mMutex);
 };
 
 // ---------------------------------------------------------------------------
@@ -164,13 +166,15 @@ XMLHttpRequestStringWriterHelper::XMLHttpRequestStringWriterHelper(
     XMLHttpRequestString& aString)
     : mBuffer(aString.mBuffer), mLock(aString.mBuffer->mMutex) {}
 
+XMLHttpRequestStringWriterHelper::~XMLHttpRequestStringWriterHelper() = default;
+
 uint32_t XMLHttpRequestStringWriterHelper::Length() const {
   return mBuffer->UnsafeLength();
 }
 
-mozilla::BulkWriteHandle<char16_t> XMLHttpRequestStringWriterHelper::BulkWrite(
-    uint32_t aCapacity, nsresult& aRv) {
-  return mBuffer->UnsafeBulkWrite(aCapacity, aRv);
+mozilla::Result<mozilla::BulkWriteHandle<char16_t>, nsresult>
+XMLHttpRequestStringWriterHelper::BulkWrite(uint32_t aCapacity) {
+  return mBuffer->UnsafeBulkWrite(aCapacity);
 }
 
 // ---------------------------------------------------------------------------
@@ -181,6 +185,9 @@ XMLHttpRequestStringSnapshotReaderHelper::
         XMLHttpRequestStringSnapshot& aSnapshot)
     : mBuffer(aSnapshot.mBuffer), mLock(aSnapshot.mBuffer->mMutex) {}
 
+XMLHttpRequestStringSnapshotReaderHelper::
+    ~XMLHttpRequestStringSnapshotReaderHelper() = default;
+
 const char16_t* XMLHttpRequestStringSnapshotReaderHelper::Buffer() const {
   return mBuffer->UnsafeData().BeginReading();
 }
@@ -189,5 +196,4 @@ uint32_t XMLHttpRequestStringSnapshotReaderHelper::Length() const {
   return mBuffer->UnsafeLength();
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
