@@ -11,28 +11,29 @@
 #include "nsXULPopupListener.h"
 #include "nsCOMPtr.h"
 #include "nsGkAtoms.h"
-#include "nsIDOMElement.h"
 #include "nsContentCID.h"
 #include "nsContentUtils.h"
 #include "nsXULPopupManager.h"
 #include "nsIScriptContext.h"
-#include "nsIDocument.h"
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/DocumentInlines.h"
 #include "nsServiceManagerUtils.h"
-#include "nsIPrincipal.h"
-#include "nsIScriptSecurityManager.h"
 #include "nsLayoutUtils.h"
 #include "mozilla/ReflowInput.h"
 #include "nsIObjectLoadingContent.h"
+#include "mozilla/BasePrincipal.h"
 #include "mozilla/EventStateManager.h"
 #include "mozilla/EventStates.h"
 #include "mozilla/Preferences.h"
-#include "mozilla/dom/Event.h"  // for nsIDOMEvent::InternalDOMEvent()
+#include "mozilla/dom/Element.h"
+#include "mozilla/dom/Event.h"  // for Event
 #include "mozilla/dom/EventTarget.h"
 #include "mozilla/dom/FragmentOrElement.h"
+#include "mozilla/dom/MouseEvent.h"
+#include "mozilla/dom/MouseEventBinding.h"
 
 // for event firing in context menus
 #include "nsPresContext.h"
-#include "nsIPresShell.h"
 #include "nsFocusManager.h"
 #include "nsPIDOMWindow.h"
 #include "nsViewManager.h"
@@ -45,7 +46,7 @@ using namespace mozilla::dom;
 // on win32 and os/2, context menus come up on mouse up. On other platforms,
 // they appear on mouse down. Certain bits of code care about this difference.
 #if defined(XP_WIN)
-#define NS_CONTEXT_MENU_IS_MOUSEUP 1
+#  define NS_CONTEXT_MENU_IS_MOUSEUP 1
 #endif
 
 nsXULPopupListener::nsXULPopupListener(mozilla::dom::Element* aElement,
@@ -85,7 +86,7 @@ NS_INTERFACE_MAP_END
 ////////////////////////////////////////////////////////////////
 // nsIDOMEventListener
 
-nsresult nsXULPopupListener::HandleEvent(nsIDOMEvent* aEvent) {
+nsresult nsXULPopupListener::HandleEvent(Event* aEvent) {
   nsAutoString eventType;
   aEvent->GetType(eventType);
 
@@ -93,77 +94,40 @@ nsresult nsXULPopupListener::HandleEvent(nsIDOMEvent* aEvent) {
         (eventType.EqualsLiteral("contextmenu") && mIsContext)))
     return NS_OK;
 
-  int16_t button;
-
-  nsCOMPtr<nsIDOMMouseEvent> mouseEvent = do_QueryInterface(aEvent);
+  MouseEvent* mouseEvent = aEvent->AsMouseEvent();
   if (!mouseEvent) {
     // non-ui event passed in.  bad things.
     return NS_OK;
   }
 
   // Get the node that was clicked on.
-  EventTarget* target = mouseEvent->AsEvent()->InternalDOMEvent()->GetTarget();
-  nsCOMPtr<nsIDOMNode> targetNode = do_QueryInterface(target);
-
-  if (!targetNode && mIsContext) {
-    // Not a DOM node, see if it's the DOM window (bug 380818).
-    nsCOMPtr<nsPIDOMWindowInner> domWin = do_QueryInterface(target);
-    if (!domWin) {
-      return NS_ERROR_DOM_WRONG_TYPE_ERR;
-    }
-    // Try to use the root node as target node.
-    nsCOMPtr<nsIDocument> doc = domWin->GetDoc();
-
-    if (doc) targetNode = do_QueryInterface(doc->GetRootElement());
-    if (!targetNode) {
-      return NS_ERROR_FAILURE;
-    }
-  }
-
-  nsCOMPtr<nsIContent> targetContent = do_QueryInterface(target);
+  nsCOMPtr<nsIContent> targetContent =
+      nsIContent::FromEventTargetOrNull(mouseEvent->GetTarget());
   if (!targetContent) {
     return NS_OK;
   }
 
-  {
-    EventTarget* originalTarget =
-        mouseEvent->AsEvent()->InternalDOMEvent()->GetOriginalTarget();
-    nsCOMPtr<nsIContent> content = do_QueryInterface(originalTarget);
-    if (content && EventStateManager::IsRemoteTarget(content)) {
+  if (nsIContent* content =
+          nsIContent::FromEventTargetOrNull(mouseEvent->GetOriginalTarget())) {
+    if (EventStateManager::IsTopLevelRemoteTarget(content)) {
       return NS_OK;
     }
   }
 
-  bool preventDefault;
-  mouseEvent->AsEvent()->GetDefaultPrevented(&preventDefault);
-  if (preventDefault && targetNode && mIsContext) {
+  bool preventDefault = mouseEvent->DefaultPrevented();
+  if (preventDefault && mIsContext) {
     // Someone called preventDefault on a context menu.
     // Let's make sure they are allowed to do so.
     bool eventEnabled =
         Preferences::GetBool("dom.event.contextmenu.enabled", true);
     if (!eventEnabled) {
-      // If the target node is for plug-in, we should not open XUL context
-      // menu on windowless plug-ins.
-      nsCOMPtr<nsIObjectLoadingContent> olc = do_QueryInterface(targetNode);
-      uint32_t type;
-      if (olc && NS_SUCCEEDED(olc->GetDisplayedType(&type)) &&
-          type == nsIObjectLoadingContent::TYPE_PLUGIN) {
-        return NS_OK;
-      }
-
       // The user wants his contextmenus.  Let's make sure that this is a
       // website and not chrome since there could be places in chrome which
       // don't want contextmenus.
-      nsCOMPtr<nsINode> node = do_QueryInterface(targetNode);
-      if (node) {
-        nsCOMPtr<nsIPrincipal> system;
-        nsContentUtils::GetSecurityManager()->GetSystemPrincipal(
-            getter_AddRefs(system));
-        if (node->NodePrincipal() != system) {
-          // This isn't chrome.  Cancel the preventDefault() and
-          // let the event go forth.
-          preventDefault = false;
-        }
+      if (!targetContent->NodePrincipal()->IsSystemPrincipal()) {
+        // This isn't chrome.  Cancel the preventDefault() and
+        // let the event go forth.
+        preventDefault = false;
       }
     }
   }
@@ -178,41 +142,38 @@ nsresult nsXULPopupListener::HandleEvent(nsIDOMEvent* aEvent) {
   // If a menu item child was clicked on that leads to a popup needing
   // to show, we know (guaranteed) that we're dealing with a menu or
   // submenu of an already-showing popup.  We don't need to do anything at all.
-  if (!mIsContext) {
-    if (targetContent &&
-        targetContent->IsAnyOfXULElements(nsGkAtoms::menu, nsGkAtoms::menuitem))
-      return NS_OK;
+  if (!mIsContext &&
+      targetContent->IsAnyOfXULElements(nsGkAtoms::menu, nsGkAtoms::menuitem)) {
+    return NS_OK;
   }
 
   if (mIsContext) {
 #ifndef NS_CONTEXT_MENU_IS_MOUSEUP
-    uint16_t inputSource = nsIDOMMouseEvent::MOZ_SOURCE_UNKNOWN;
-    mouseEvent->GetMozInputSource(&inputSource);
-    bool isTouch = inputSource == nsIDOMMouseEvent::MOZ_SOURCE_TOUCH;
+    uint16_t inputSource = mouseEvent->MozInputSource();
+    bool isTouch = inputSource == MouseEvent_Binding::MOZ_SOURCE_TOUCH;
     // If the context menu launches on mousedown,
     // we have to fire focus on the content we clicked on
-    FireFocusOnTargetContent(targetNode, isTouch);
+    FireFocusOnTargetContent(targetContent, isTouch);
 #endif
   } else {
     // Only open popups when the left mouse button is down.
-    mouseEvent->GetButton(&button);
-    if (button != 0) return NS_OK;
+    if (mouseEvent->Button() != 0) {
+      return NS_OK;
+    }
   }
 
   // Open the popup. LaunchPopup will call StopPropagation and PreventDefault
   // in the right situations.
-  LaunchPopup(aEvent, targetContent);
+  LaunchPopup(mouseEvent);
 
   return NS_OK;
 }
 
 #ifndef NS_CONTEXT_MENU_IS_MOUSEUP
-nsresult nsXULPopupListener::FireFocusOnTargetContent(nsIDOMNode* aTargetNode,
-                                                      bool aIsTouch) {
-  nsCOMPtr<nsIContent> content = do_QueryInterface(aTargetNode);
-  nsCOMPtr<nsIDocument> doc = content->OwnerDoc();
+nsresult nsXULPopupListener::FireFocusOnTargetContent(
+    nsIContent* aTargetContent, bool aIsTouch) {
+  nsCOMPtr<Document> doc = aTargetContent->OwnerDoc();
 
-  // Get nsIDOMElement for targetNode
   // strong reference to keep this from going away between events
   // XXXbz between what events?  We don't use this local at all!
   RefPtr<nsPresContext> context = doc->GetPresContext();
@@ -220,48 +181,41 @@ nsresult nsXULPopupListener::FireFocusOnTargetContent(nsIDOMNode* aTargetNode,
     return NS_ERROR_FAILURE;
   }
 
-  nsIFrame* targetFrame = content->GetPrimaryFrame();
+  nsIFrame* targetFrame = aTargetContent->GetPrimaryFrame();
   if (!targetFrame) return NS_ERROR_FAILURE;
 
-  const nsStyleUserInterface* ui = targetFrame->StyleUserInterface();
-  bool suppressBlur = (ui->mUserFocus == StyleUserFocus::Ignore);
+  const bool suppressBlur =
+      targetFrame->StyleUI()->UserFocus() == StyleUserFocus::Ignore;
 
-  nsCOMPtr<nsIDOMElement> element;
-  nsCOMPtr<nsIContent> newFocus = content;
+  RefPtr<Element> newFocusElement;
 
   nsIFrame* currFrame = targetFrame;
   // Look for the nearest enclosing focusable frame.
   while (currFrame) {
-    int32_t tabIndexUnused;
-    if (currFrame->IsFocusable(&tabIndexUnused, true)) {
-      newFocus = currFrame->GetContent();
-      nsCOMPtr<nsIDOMElement> domElement(do_QueryInterface(newFocus));
-      if (domElement) {
-        element = domElement;
-        break;
-      }
+    if (currFrame->IsFocusable(/* aWithMouse = */ true) &&
+        currFrame->GetContent()->IsElement()) {
+      newFocusElement = currFrame->GetContent()->AsElement();
+      break;
     }
     currFrame = currFrame->GetParent();
   }
 
-  nsIFocusManager* fm = nsFocusManager::GetFocusManager();
-  if (fm) {
-    if (element) {
+  if (RefPtr<nsFocusManager> fm = nsFocusManager::GetFocusManager()) {
+    if (newFocusElement) {
       uint32_t focusFlags =
           nsIFocusManager::FLAG_BYMOUSE | nsIFocusManager::FLAG_NOSCROLL;
       if (aIsTouch) {
         focusFlags |= nsIFocusManager::FLAG_BYTOUCH;
       }
-      fm->SetFocus(element, focusFlags);
+      fm->SetFocus(newFocusElement, focusFlags);
     } else if (!suppressBlur) {
-      nsPIDOMWindowOuter* window = doc->GetWindow();
+      nsCOMPtr<nsPIDOMWindowOuter> window = doc->GetWindow();
       fm->ClearFocus(window);
     }
   }
 
   EventStateManager* esm = context->EventStateManager();
-  nsCOMPtr<nsIContent> focusableContent = do_QueryInterface(element);
-  esm->SetContentState(focusableContent, NS_EVENT_STATE_ACTIVE);
+  esm->SetContentState(newFocusElement, NS_EVENT_STATE_ACTIVE);
 
   return NS_OK;
 }
@@ -312,8 +266,7 @@ static already_AddRefed<Element> GetImmediateChild(nsIContent* aContent,
 // (popup, context) and uses that attribute's value as an ID for
 // the popup content in the document.
 //
-nsresult nsXULPopupListener::LaunchPopup(nsIDOMEvent* aEvent,
-                                         nsIContent* aTargetContent) {
+nsresult nsXULPopupListener::LaunchPopup(MouseEvent* aEvent) {
   nsresult rv = NS_OK;
 
   nsAutoString identifier;
@@ -336,7 +289,7 @@ nsresult nsXULPopupListener::LaunchPopup(nsIDOMEvent* aEvent,
   if (identifier.IsEmpty()) return rv;
 
   // Try to find the popup content and the document.
-  nsCOMPtr<nsIDocument> document = mElement->GetComposedDoc();
+  nsCOMPtr<Document> document = mElement->GetComposedDoc();
   if (!document) {
     NS_WARNING("No document!");
     return NS_ERROR_FAILURE;
@@ -346,20 +299,6 @@ nsresult nsXULPopupListener::LaunchPopup(nsIDOMEvent* aEvent,
   RefPtr<Element> popup;
   if (identifier.EqualsLiteral("_child")) {
     popup = GetImmediateChild(mElement, nsGkAtoms::menupopup);
-    if (!popup) {
-      nsINodeList* list = document->GetAnonymousNodes(*mElement);
-      if (list) {
-        uint32_t listLength = list->Length();
-        for (uint32_t ctr = 0; ctr < listLength; ctr++) {
-          nsIContent* childContent = list->Item(ctr);
-          if (childContent->NodeInfo()->Equals(nsGkAtoms::menupopup,
-                                               kNameSpaceID_XUL)) {
-            popup = childContent->AsElement();
-            break;
-          }
-        }
-      }
-    }
   } else if (!mElement->IsInUncomposedDoc() ||
              !(popup = document->GetElementById(identifier))) {
     // XXXsmaug Should we try to use ShadowRoot::GetElementById in case
@@ -394,15 +333,11 @@ nsresult nsXULPopupListener::LaunchPopup(nsIDOMEvent* aEvent,
       (mPopupContent->HasAttr(kNameSpaceID_None, nsGkAtoms::position) ||
        (mPopupContent->HasAttr(kNameSpaceID_None, nsGkAtoms::popupanchor) &&
         mPopupContent->HasAttr(kNameSpaceID_None, nsGkAtoms::popupalign)))) {
-    pm->ShowPopup(mPopupContent, mElement, EmptyString(), 0, 0, false, true,
-                  false, aEvent);
+    pm->ShowPopup(mPopupContent, mElement, u""_ns, 0, 0, false, true, false,
+                  aEvent);
   } else {
-    int32_t xPos = 0, yPos = 0;
-    nsCOMPtr<nsIDOMMouseEvent> mouseEvent = do_QueryInterface(aEvent);
-    mouseEvent->GetScreenX(&xPos);
-    mouseEvent->GetScreenY(&yPos);
-
-    pm->ShowPopupAtScreen(mPopupContent, xPos, yPos, mIsContext, aEvent);
+    CSSIntPoint pos = aEvent->ScreenPoint(CallerType::System);
+    pm->ShowPopupAtScreen(mPopupContent, pos.x, pos.y, mIsContext, aEvent);
   }
 
   return NS_OK;
