@@ -21,7 +21,6 @@ class PresShell;
 
 namespace mozilla::layers {
 class Layer;
-class RefLayer;
 class RenderRootStateManager;
 class WebRenderLayerScrollData;
 class WebRenderScrollData;
@@ -110,6 +109,7 @@ class nsSubDocumentFrame final : public nsAtomicContainerFrame,
   nsIFrame* GetSubdocumentRootFrame();
   enum { IGNORE_PAINT_SUPPRESSION = 0x1 };
   mozilla::PresShell* GetSubdocumentPresShellForPainting(uint32_t aFlags);
+  nsRect GetDestRect();
   mozilla::ScreenIntSize GetSubdocumentSize();
 
   // nsIReflowCallback
@@ -129,8 +129,12 @@ class nsSubDocumentFrame final : public nsAtomicContainerFrame,
   }
 
   nsFrameLoader* FrameLoader() const;
-  void ResetFrameLoader();
 
+  enum class RetainPaintData : bool { No, Yes };
+  void ResetFrameLoader(RetainPaintData);
+  void ClearRetainedPaintData();
+
+  void MaybeUpdateEmbedderColorScheme();
   void PropagateIsUnderHiddenEmbedderElementToSubView(
       bool aIsUnderHiddenEmbedderElement);
 
@@ -138,6 +142,14 @@ class nsSubDocumentFrame final : public nsAtomicContainerFrame,
 
   void SubdocumentIntrinsicSizeOrRatioChanged();
 
+  struct RemoteFramePaintData {
+    mozilla::layers::LayersId mLayersId;
+    mozilla::dom::TabId mTabId{0};
+  };
+
+  RemoteFramePaintData GetRemotePaintData() const;
+  bool HasRetainedPaintData() const { return mRetainedRemoteFrame.isSome(); }
+
  protected:
   friend class AsyncFrameInit;
 
@@ -151,6 +163,9 @@ class nsSubDocumentFrame final : public nsAtomicContainerFrame,
   }
 
   nscoord GetIntrinsicISize() {
+    if (StyleDisplay()->GetContainSizeAxes().mIContained) {
+      return 0;
+    }
     auto size = GetIntrinsicSize();
     Maybe<nscoord> iSize =
         GetWritingMode().IsVertical() ? size.height : size.width;
@@ -170,24 +185,28 @@ class nsSubDocumentFrame final : public nsAtomicContainerFrame,
   nsView* mOuterView;
   nsView* mInnerView;
 
-  bool mIsInline;
-  bool mPostedReflowCallback;
-  bool mDidCreateDoc;
-  bool mCallingShow;
+  // When process-switching a remote tab, we might temporarily paint the old
+  // one.
+  Maybe<RemoteFramePaintData> mRetainedRemoteFrame;
+
+  bool mIsInline : 1;
+  bool mPostedReflowCallback : 1;
+  bool mDidCreateDoc : 1;
+  bool mCallingShow : 1;
 };
 
+namespace mozilla {
+
 /**
  * A nsDisplayRemote will graft a remote frame's shadow layer tree (for a given
  * nsFrameLoader) into its parent frame's layer tree.
  */
 class nsDisplayRemote final : public nsPaintedDisplayItem {
-  typedef mozilla::ContainerLayerParameters ContainerLayerParameters;
   typedef mozilla::dom::TabId TabId;
   typedef mozilla::gfx::Matrix4x4 Matrix4x4;
   typedef mozilla::layers::EventRegionsOverride EventRegionsOverride;
   typedef mozilla::layers::Layer Layer;
   typedef mozilla::layers::LayersId LayersId;
-  typedef mozilla::layers::RefLayer RefLayer;
   typedef mozilla::layers::StackingContextHelper StackingContextHelper;
   typedef mozilla::LayerState LayerState;
   typedef mozilla::LayoutDeviceRect LayoutDeviceRect;
@@ -196,14 +215,6 @@ class nsDisplayRemote final : public nsPaintedDisplayItem {
  public:
   nsDisplayRemote(nsDisplayListBuilder* aBuilder, nsSubDocumentFrame* aFrame);
 
-  LayerState GetLayerState(
-      nsDisplayListBuilder* aBuilder, LayerManager* aManager,
-      const ContainerLayerParameters& aParameters) override;
-
-  already_AddRefed<Layer> BuildLayer(
-      nsDisplayListBuilder* aBuilder, LayerManager* aManager,
-      const ContainerLayerParameters& aContainerParameters) override;
-
   void Paint(nsDisplayListBuilder* aBuilder, gfxContext* aCtx) override;
 
   bool CreateWebRenderCommands(
@@ -219,13 +230,16 @@ class nsDisplayRemote final : public nsPaintedDisplayItem {
   NS_DISPLAY_DECL_NAME("Remote", TYPE_REMOTE)
 
  private:
-  friend class nsDisplayItemBase;
+  friend class nsDisplayItem;
+  using RemoteFramePaintData = nsSubDocumentFrame::RemoteFramePaintData;
+
   nsFrameLoader* GetFrameLoader() const;
 
-  TabId mTabId;
-  LayersId mLayersId;
+  RemoteFramePaintData mPaintData;
   LayoutDevicePoint mOffset;
   EventRegionsOverride mEventRegionsOverride;
 };
 
+}  // namespace mozilla
+
 #endif /* NSSUBDOCUMENTFRAME_H_ */
