@@ -6,8 +6,11 @@
 
 #include "nsMathMLmmultiscriptsFrame.h"
 
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/Element.h"
 #include "mozilla/PresShell.h"
 #include "mozilla/StaticPrefs_mathml.h"
+#include "nsLayoutUtils.h"
 #include "nsPresContext.h"
 #include <algorithm>
 #include "gfxContext.h"
@@ -113,7 +116,7 @@ nsresult nsMathMLmmultiscriptsFrame::Place(DrawTarget* aDrawTarget,
         mContent->AsElement()->GetAttr(kNameSpaceID_None,
                                        nsGkAtoms::subscriptshift_, value)) {
       mContent->OwnerDoc()->WarnOnceAbout(
-          dom::Document::eMathML_DeprecatedScriptShiftAttributes);
+          dom::DeprecatedOperations::eMathML_DeprecatedScriptShiftAttributes);
       ParseNumericValue(value, &subScriptShift, 0, PresContext(),
                         mComputedStyle, fontSizeInflation);
     }
@@ -132,7 +135,7 @@ nsresult nsMathMLmmultiscriptsFrame::Place(DrawTarget* aDrawTarget,
         mContent->AsElement()->GetAttr(kNameSpaceID_None,
                                        nsGkAtoms::superscriptshift_, value)) {
       mContent->OwnerDoc()->WarnOnceAbout(
-          dom::Document::eMathML_DeprecatedScriptShiftAttributes);
+          dom::DeprecatedOperations::eMathML_DeprecatedScriptShiftAttributes);
       ParseNumericValue(value, &supScriptShift, 0, PresContext(),
                         mComputedStyle, fontSizeInflation);
     }
@@ -267,8 +270,8 @@ nsresult nsMathMLmmultiscriptsFrame::PlaceMultiScript(
 
     // get sup script shift depending on current script level and display style
     // Rule 18c, App. G, TeXbook
-    if (font->mScriptLevel == 0 &&
-        font->mMathDisplay == NS_MATHML_DISPLAYSTYLE_BLOCK &&
+    if (font->mMathDepth == 0 &&
+        font->mMathStyle == NS_STYLE_MATH_STYLE_NORMAL &&
         !NS_MATHML_IS_COMPRESSED(presentationData.flags)) {
       // Style D in TeXbook
       supScriptShift = supScriptShift1;
